package com.sillycycle.bagleyd.barrel;

/*
 * @(#)BarrelFrame.java
 *
 * Copyright 2004 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;

/**
 * The <code>BarrelFrame</code> class is to be run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class BarrelFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	//static final int MAX_TILES = 8;
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int NEVER = -1;
	static final String SCOREFILE = "barrel.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Undo", "Redo", "Clear",
		"Practice", "Randomize", "Auto-solve",
		"Number", "Orientize", "Pairs",
		"Speed >", "Slow <", "Sound @",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*',
		'u', 'r', 'c', 'p', 'z', 's',
		'n', 'o', 'i', '>', '<', '@',
		'?', '!', '^', 'a', 'G', 'W', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 512;
	int heightValue = 512;
	static int xValue = 0;
	static int yValue = 0;
	String monoString = "GRYOBK";
	Color[] faceColor = new Color[] {
		Color.green, Color.red, Color.yellow, Color.orange,
		Color.blue, Color.black
	};
	Color tileColor = Color.lightGray;
	Color borderColor = Color.darkGray;
	Color fgColor = Color.cyan;
	Color bgColor = Color.gray;
	boolean sound = false;
	boolean controlValue = true;
	boolean orientValue = BarrelInterface.DEFAULT_ORIENT;
	boolean pairsValue = BarrelInterface.DEFAULT_PAIRS;
	boolean practiceValue = BarrelInterface.DEFAULT_PRACTICE;
	int tileValue = BarrelInterface.DEFAULT_TILES;
	int faceValue = BarrelInterface.DEFAULT_FACES;
	int baseValue = BarrelInterface.DEFAULT_BASE;
	int delayValue = 10;
	int speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
	boolean debug = false;
	JPanel topPanel = new JPanel();
	JCheckBox orientCheckBox = new JCheckBox("Oriented");
	JCheckBox pairsCheckBox = new JCheckBox("Pairs");
	JCheckBox practiceCheckBox = new JCheckBox("Practice");
	Barrel puzzle;
	int movesDisplay = 0;
	String recordDisplay = "NEVER";
	String messageDisplay = "Play Barrel! (use mouse or keypad)";
	JTextField movesValueLabel = new JTextField(Integer.toString(
		movesDisplay));
	JTextField recordValueLabel = new JTextField(recordDisplay);
	Label messageLabel = new Label(messageDisplay, Label.CENTER);
	// JLabel constantly resizes
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	IODialog ioDialog = null;
	int[][] puzzleRecord = null;
	JMenuBar menuBar;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	//JCheckBoxMenuItem numberMenuItem = null;
        JCheckBoxMenuItem soundMenuItem = null;
	Spinner speedChanger = null;
	//Slider tileChanger = null, faceChanger = null, speedChanger = null;

	public String writePuzzleString() {
		return this.stringBuilderSave.toString();
	}

	public void getPuzzleString(String s) {
		puzzle.stringSave = s;
	}

	public void processKey(char key, boolean shift) {
		puzzle.processKey(key, shift);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
		if (controlValue)
			soundMenuItem.setSelected(sound);
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		for (int i = 0; i < faceColor.length; i++)
			faceColor[i] = getOpt("faceColor" + i, faceColor[i]);
		tileColor = getOpt("tileColor", tileColor);
		borderColor = getOpt("borderColor", borderColor);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		sound = getOpt("sound", sound);
		controlValue = getOpt("control", controlValue);
		/*tileValue = getOpt("tiles", tileValue);*/
		if (tileValue < BarrelInterface.MIN_TILES)
			tileValue = BarrelInterface.DEFAULT_TILES;
		/*faceValue = getOpt("faces", faceValue);*/
		if (faceValue < BarrelInterface.MIN_FACES)
			faceValue = BarrelInterface.DEFAULT_FACES;
		orientValue = getOpt("orient", orientValue);
		pairsValue = getOpt("pairs", pairsValue);
		practiceValue = getOpt("practice", practiceValue);
		baseValue = getOpt("base", baseValue);
		if (baseValue < BarrelInterface.MIN_BASE ||
				baseValue > BarrelInterface.MAX_BASE)
			baseValue = BarrelInterface.DEFAULT_BASE;
		delayValue = getOpt("delay", delayValue);
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Barrel", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Barrel - The Barrel, the Billion Barrel, or Wonderful Barrel also" + NL +
"seen as the Cylinder Puzzle is of Japanese origin.  The puzzle" + NL +
"contains 23 beads of 6 different colors.  4 each of 5 different" + NL +
"colors arranged in 5 rows and 3 black beads making 3 rows one unit" + NL +
"unit longer.  A system of plugers slide all three of the long rows" + NL +
"left and right.  Two of the plungers sit together and the other" + NL +
"(the middle plunger) is separate from the other two.  There are" + NL +
"2 separate wheels or disks that rotate about the center axis." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Make sure you randomize the puzzle first before trying to move the" + NL +
"\"tiles\" for the first time." + NL +
" " + NL +
"Press \"mouse-left\" button to move a tile.  Release \"mouse-left\"" + NL +
"button in the same row would move the tile towards the space" + NL +
"(unless there is no space in that row).  Release on a different" + NL +
"row (and the same column) will move the tiles up or down." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"Z\" or \"z\" keys, to" + NL +
"randomize the puzzle.  One must double click on \"mouse-" + NL +
"right\" if the puzzle is currently being worked on." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved puzzle to screen." + NL +
" " + NL +
"Press \"w\" key to save (write) a puzzle to screen." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved puzzle to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a puzzle to file." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo a move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo a move." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear the puzzle." + NL +
" " + NL +
"Press \"P\" or \"p\" keys, to toggle the practice mode (in practice" + NL +
"mode the record should say \"practice\").  This is good for" + NL +
"learning moves and experimenting." + NL +
" " + NL +
"Press \"S\" or \"s\" keys to start the auto-solver.  Only works " + NL +
"where oriented is off." + NL +
" " + NL +
"Press \"N\" or \"n\" keys to toggle the numbers or non-graphics mode." + NL +
"(Default is on)." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to toggle the orient mode.  (Default is off)." + NL +
"If on, puzzle has to solved in numerical order." + NL +
" " + NL +
"Press \"I\" or \"i\" keys to toggle pairs mode.  (Default is on)." + NL +
"If on, all the tiles rotate in pairs." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of the tiles." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow up the movement of the tiles." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL +
" " + NL +
"Use the key pad or arrow keys to move without the" + NL +
"mouse." + NL +
"Key pad (better with Num Lock on) is defined for Barrel as:" + NL +
"    8   Up" + NL +
"    ^" + NL +
"4< >6  Left, Right" + NL +
"    v" + NL +
"    2   Down" + NL +
"The arrow keys seem at first to work in the opposite direction" + NL +
"then you expect because the spaces are not the objects moving." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"L. E. Horden, Sliding Piece Puzzles (Recreations in " +
"Mathematics Series), Oxford University" + NL +
"Press 1986, Plate X." + NL +
"James G Nourse, The Simple Solutions to Cubic, Puzzles, " +
"Bantam Books, New York, November 1981, pp 25-42." + NL +
"John Ewing & Czes Kosniowski, Puzzle it Out: " +
"Cubes Groups and Puzzles, Cambridge University Press," + NL +
"New York, 1982, pp 50-54?" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java BarrelFrame Version 8.7.2" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/puzzles.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = BarrelFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = BarrelFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playBumpAudio() {
		playClip(BarrelInterface.SOUNDS[BarrelInterface.Sounds.BUMP.ordinal()]);
	}

	public void playMoveAudio() {
		playClip(BarrelInterface.SOUNDS[BarrelInterface.Sounds.MOVE.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (puzzle == null)
			return;
		puzzle.setBounds(0, 0, width, height);
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		puzzle.setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		StringBuilder scoreString = new StringBuilder();

		for (int i = 0; i < puzzleRecord.length; i++) {
			for (int j = 0; j < puzzleRecord[i].length; j++) {
				scoreString.append(puzzleRecord[i][j]);
				if (j < puzzleRecord[i].length - 1)
					scoreString.append(" ");
			}
			if (i < puzzleRecord.length - 1)
				scoreString.append(NL);
		}
		return scoreString.toString();
	}

	public void writeScore() {
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(readUserScore());
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(String cookieValue, boolean store) {
		String[] st = null;
		int field = 0;

		if (cookieValue != null)
			st = cookieValue.split("[+/\\s]+");
		if (debug)
			System.out.println("initRecords: " + cookieValue);
		puzzleRecord = new int[2][2];
		for (int i = 0; i < puzzleRecord.length; i++) {
		    for (int j = 0; j < puzzleRecord[i].length; j++) {
			if (st == null || st.length <= field) {
				//build it from the array
				puzzleRecord[i][j] =
					BarrelScores.PUZZLE_SCORES[i][j];
					//MAXRECORD;
			} else {
				puzzleRecord[i][j] =
					Integer.parseInt(st[field++]);
			}
		    }
		}
		if (store && st == null)
			writeScore(); // if no cookie, then make one...
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	void printRecord(boolean orient, boolean pairs,
			boolean practice) {
		int i = (pairs) ? 1 : 0;
		int j = (orient) ? 1 : 0;

		if (puzzleRecord == null)
			return;
		if (practice) {
			recordValueLabel.setText("PRACTICE");
		/*} else if (tileValue > MAX_TILES) {
			recordValueLabel.setText("NONE");*/
		} else if (puzzleRecord[i][j] <= NEVER) {
			recordValueLabel.setText("NEVER");
		} else {
			recordValueLabel.setText(Integer.toString(
				puzzleRecord[i][j]));
		}
	}

	boolean handleSolved(int counter, boolean orient,
			boolean pairs, boolean practice) {
		int i = (pairs) ? 1 : 0;
		int j = (orient) ? 1 : 0;

		if (puzzleRecord != null &&
				/*tileValue <= MAX_TILES &&*/
				(counter < puzzleRecord[i][j] ||
				puzzleRecord[i][j] <= NEVER)) {
			readScoreRecords(); /* Maybe its been updated by another */
			puzzleRecord[i][j] = counter;
			writeScore();
			printRecord(orient, pairs,
				practice);
			return true;
		}
		return false;
	}

	public void initScore() {
		if (puzzleRecord == null) {
			readScoreRecords();
			printRecord(orientValue, pairsValue,
				practiceValue);
		}
	}

	/*class TileChangeListener implements ChangeListener {
		public void stateChanged(ChangeEvent event) {
			int oldValue = tileValue;

			tileValue = tileChanger.getValue();
			if (oldValue != tileValue) {
				puzzle.setPuzzleTiles(tileValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(orientValue, pairsValue,
					practiceValue);
			}
		}
	}

	class FaceChangeListener implements ChangeListener {
		public void stateChanged(ChangeEvent event) {
			int oldValue = faceValue;

			faceValue = faceChanger.getValue();
			if (oldValue != faceValue) {
				puzzle.setPuzzleFaces(faceValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(orientValue, pairsValue,
					practiceValue);
			}
		}
	}*/

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				puzzle.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	class OrientItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = orientValue;

			if (!puzzle.getNumber()) {
				orientCheckBox.setSelected(oldValue);
				return;
			}
			orientValue = orientCheckBox.isSelected();
			if (oldValue != orientValue) {
				puzzle.setPuzzleOrient(orientValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(orientValue, pairsValue,
					practiceValue);
			}
		}
	}

	class PairsItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = pairsValue;

			pairsValue = pairsCheckBox.isSelected();
			if (oldValue != pairsValue) {
				puzzle.setPuzzlePairs(pairsValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(orientValue, pairsValue,
					practiceValue);
			}
		}
	}

	class PracticeItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = practiceValue;

			practiceValue = practiceCheckBox.isSelected();
			if (oldValue != practiceValue) {
				puzzle.setPuzzlePractice(practiceValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(orientValue, pairsValue,
					practiceValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt],
					((event.getModifiers() &
					InputEvent.SHIFT_DOWN_MASK) != 0));
				return;
			}
		}
	}

	public void callback(int reason) {
		messageDisplay = "										";
		switch(reason) {
		case BarrelInterface.ACTION_RESTORE:
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			movesDisplay = 0;
			break;
		case BarrelInterface.ACTION_RESET:
			movesDisplay = 0;
			break;
		case BarrelInterface.ACTION_BLOCKED:
			messageDisplay = "Blocked";
			break;
		case BarrelInterface.ACTION_SPACE:
			// Too annoying
			// messageDisplay = "A space cannot slide";
			break;
		case BarrelInterface.ACTION_IGNORE:
			messageDisplay = "Randomize to start";
			break;
		case BarrelInterface.ACTION_MOVED:
			movesDisplay++;
			break;
		case BarrelInterface.ACTION_SOLVED:
			if (practiceValue)
				movesDisplay = 0;
			else {
				if (puzzle.getCheat())
					messageDisplay = "No cheating!!"; //username
				else if (handleSolved(movesDisplay, orientValue,
						pairsValue, practiceValue))
					messageDisplay =
						"Best solution, congratulations!!"; //username
				else
					messageDisplay = "Solved!";
			}
			//showMessage(messageDisplay);
			break;
		case BarrelInterface.ACTION_PRACTICE:
			movesDisplay = 0;
			practiceValue = !practiceValue;
			if (!practiceValue)
				messageDisplay = "Randomize to start";
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzlePractice(practiceValue);
			practiceCheckBox.setSelected(practiceValue);
			break;
		case BarrelInterface.ACTION_RANDOMIZE:
			movesDisplay = 0;
			practiceValue = false;
			puzzle.setPuzzlePractice(practiceValue);
			break;
		case BarrelInterface.ACTION_COMPUTED:
			break;
		case BarrelInterface.ACTION_ORIENTIZE:
			movesDisplay = 0;
			orientValue = !orientValue;
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzleOrient(orientValue);
			orientCheckBox.setSelected(orientValue);
			break;
		case BarrelInterface.ACTION_PAIRS:
			movesDisplay = 0;
			pairsValue = !pairsValue;
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzlePairs(pairsValue);
			pairsCheckBox.setSelected(pairsValue);
			break;
		/*case BarrelInterface.ACTION_DECX:
			movesDisplay = 0;
			tileValue--;
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzleTiles(tileValue);
			if (tileChanger != null)
				tileChanger.setValue(tileValue);
			break;
		case BarrelInterface.ACTION_INCX:
			movesDisplay = 0;
			tileValue++;
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzleTiles(tileValue);
			if (tileChanger != null)
				tileChanger.setValue(tileValue);
			break;
		case BarrelInterface.ACTION_DECY:
			movesDisplay = 0;
			faceValue--;
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzleFaces(faceValue);
			if (faceValue <= BarrelInterface.MAX_FACES &&
					faceChanger != null)
				faceChanger.setValue(faceValue);
			break;
		case BarrelInterface.ACTION_INCY:
			movesDisplay = 0;
			faceValue++;
			printRecord(orientValue, pairsValue,
				practiceValue);
			puzzle.setPuzzleFaces(faceValue);
			if (faceValue <= BarrelInterface.MAX_FACES &&
					faceChanger != null)
				faceChanger.setValue(faceValue);
			break;*/
		case BarrelInterface.ACTION_UNDO:
			movesDisplay--;
			break;
		case BarrelInterface.ACTION_REDO:
			movesDisplay++;
			break;
		case BarrelInterface.ACTION_SPEED:
			{
				int oldDelay, delay;

				delay = puzzle.getDelay();
				oldDelay = delay;
				if (delay > MAX_SPEED - MIN_SPEED)
					delay = MAX_SPEED - MIN_SPEED;
				if (delay < 0)
					delay = 0;
				if (delay != oldDelay) {
					puzzle.setDelay(delay);
				}
				if (speedChanger != null)
					speedChanger.setValue(MAX_SPEED +
						MIN_SPEED - 1 - delay);
			}
			return;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(movesDisplay));
	}

	public BarrelFrame(String title) {
		/*JPanel info1Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel movesPanel = new JPanel();
		JPanel recordPanel = new JPanel();
		JPanel checkBoxPanel = new JPanel();
		/*Panel sliderPanel = new JPanel();*/
		JPanel canvasPanel = new JPanel();

		try {
			mediumIcon = getIcon(BarrelInterface.ICONS[BarrelInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(BarrelInterface.ICONS[BarrelInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			BarrelInterface.TITLE + ": IO", "Puzzle Position",
			"DONE");
		readScoreRecords();
		getOpts();
		speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Undo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_U);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Practice");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Randomize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Z);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Auto-solve");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Number");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_N);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Orientize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_O);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Pairs");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_I);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			soundMenuItem = new JCheckBoxMenuItem("Sound @", sound);
			menu.add(soundMenuItem);
			soundMenuItem.addActionListener(this);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			leftPanel.add(new JLabel("Move tile"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Randomize"));
			leftPanel.add(getIcon("mouse-l.png", 16, 16));
			rightPanel.add(getIcon("mouse-r.png", 16, 16));
			info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info1Panel.add(leftPanel);
			info1Panel.add(rightPanel);*/
			movesPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			movesPanel.add(new JLabel("Moves"));
			movesValueLabel.setColumns(5);
			movesValueLabel.setEditable(false);
			movesPanel.add(movesValueLabel);
			recordPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			recordPanel.add(new JLabel("Record"));
			recordValueLabel.setColumns(8);
			recordValueLabel.setEditable(false);
			recordPanel.add(recordValueLabel);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 5));
			info2Panel.add(movesPanel);
			info2Panel.add(recordPanel);
			/*sliderPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			tileChanger = new Slider("Tiles",
				tileValue, BarrelInterface.MIN_TILES,
				((tileValue > MAX_TILES) ? tileValue : MAX_TILES));
			sliderPanel.add(tileChanger);
			faceChanger = new Slider("Faces",
				faceValue, BarrelInterface.MIN_FACES,
				((faceValue > BarrelInterface.MAX_FACES) ? faceValue :
				BarrelInterface.MAX_FACES));
			sliderPanel.add(faceChanger);*/
			checkBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			checkBoxPanel.add(orientCheckBox);
			orientCheckBox.setSelected(orientValue);
			checkBoxPanel.add(pairsCheckBox);
			pairsCheckBox.setSelected(pairsValue);
			checkBoxPanel.add(practiceCheckBox);
			practiceCheckBox.setSelected(practiceValue);
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			topPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			/*topPanel.add(info1Panel);*/
			topPanel.add(info2Panel);
			/*topPanel.add(sliderPanel);*/
			speedChanger = new Spinner("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			/*speedChanger = new Slider("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			speedChanger.setWidth(150);*/
			topPanel.add(speedChanger);
			topPanel.add(checkBoxPanel);
			topPanel.add(messageLabel);
			startPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			startPanel.add(topPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		}
		puzzle = new Barrel(this,
			orientValue, pairsValue, practiceValue,
			tileValue, faceValue,
			baseValue, false, false, monoString,
			faceColor[0], faceColor[1], faceColor[2],
			faceColor[3], faceColor[4], faceColor[5],
			tileColor, borderColor,
			fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		canvasPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvasPanel.add(puzzle);
		secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			printRecord(orientValue, pairsValue, practiceValue);
			/*tileChanger.addChangeListener(new TileChangeListener());
			faceChanger.addChangeListener(new FaceChangeListener());*/
			speedChanger.addChangeListener(new SpeedChangeListener());
			orientCheckBox.addItemListener(new OrientItemListener());
			pairsCheckBox.addItemListener(new PairsItemListener());
			practiceCheckBox.addItemListener(new PracticeItemListener());
		}
		canvasPanel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle.requestFocus();
			}
		});
		setTitle(title);
		/* puzzle.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : BarrelInterface.ICONS) {
			URL url = BarrelFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
//"\t[-tiles={int}] [-faces={int}]"
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-faceColor{0|1|2|3|4|5}={int}]" + NL +
"\t[-tileColor={int}] [-borderColor={int}]" + NL +
"\t[-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-control={bool}]" + NL +
"\t[-orient={bool}] [-pairs={bool}] [-practice={bool}]" + NL +
"\t[-delay={int}]");
			return;
		}
		BarrelFrame frame = new BarrelFrame(BarrelInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.puzzle.requestFocusInWindow();
		frame.setLocation(BarrelFrame.xValue, BarrelFrame.yValue);
		frame.setVisible(true);
	}
}
